@@ -5,7 +5,7 @@
 #include "libcgo.h"
 
 /* Stub for calling malloc from Go */
-static void
+void
 x_cgo_malloc(void *p)
 {
 	struct a {
@@ -16,10 +16,8 @@ x_cgo_malloc(void *p)
 	a->ret = malloc(a->n);
 }
 
-void (*_cgo_malloc)(void*) = x_cgo_malloc;
-
 /* Stub for calling free from Go */
-static void
+void
 x_cgo_free(void *p)
 {
 	struct a {
@@ -29,11 +27,9 @@ x_cgo_free(void *p)
 	free(a->arg);
 }
 
-void (*_cgo_free)(void*) = x_cgo_free;
-
 /* Stub for creating a new thread */
-static void
-xlibcgo_thread_start(ThreadStart *arg)
+void
+x_cgo_thread_start(ThreadStart *arg)
 {
 	ThreadStart *ts;
 
@@ -45,7 +41,5 @@ xlibcgo_thread_start(ThreadStart *arg)
 	}
 	*ts = *arg;
 
-	libcgo_sys_thread_start(ts);	/* OS-dependent half */
+	_cgo_sys_thread_start(ts);	/* OS-dependent half */
 }
-
-void (*libcgo_thread_start)(ThreadStart*) = xlibcgo_thread_start;
